/* server.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "server.h"
#include "player.h"

#include <algorithm>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h> 
#include <errno.h>

Server::
Server(const string &sound_device, int port) {
  _sound_device = sound_device;
  _active_player = NULL;
  _exit_flag = false;

  struct sockaddr_in serv_addr;

  // open an Internet stream socket
  _welcome_socket = socket(AF_INET, SOCK_STREAM, 0);
  if (_welcome_socket < 0) {
    perror("socket");
    exit(1);
  }
  
  // write specified # of null bytes to server socket structure
  memset((char *)&serv_addr, 0, sizeof(serv_addr));
  
  // fill in socket structure
  serv_addr.sin_family      = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY); 
  serv_addr.sin_port        = htons(port);
  
  // bind local address so a client can contact us
  int bind_result = 
    bind(_welcome_socket, (struct sockaddr *)&serv_addr, sizeof(serv_addr));
  if (bind_result < 0) {
    perror("bind");
    exit(1);
  }  
  
  // set up listening queue for connection requests
  listen(_welcome_socket, 5);
}

Server::
~Server() {
  // Shut down everything on exit.
  Connections::iterator ci;
  for (ci = _closed_connections.begin(); 
       ci != _closed_connections.end(); 
       ++ci) {
    delete (*ci).second;
  }
  for (ci = _connections.begin(); ci != _connections.end(); ++ci) {
    delete (*ci).second;
  }

  close(_welcome_socket);
}
  

void Server::
main_loop() {
  while (!_exit_flag) {
    // First, delete any connections that got closed last time through
    // the loop.
    Connections::iterator ci;
    for (ci = _closed_connections.begin(); 
	 ci != _closed_connections.end(); 
	 ++ci) {
      if (_active_player == (*ci).second) {
	_active_player = NULL;
      }
      delete (*ci).second;
      _connections.erase((*ci).first);
    }
    _closed_connections.clear();

    // Perform a select() to wait for input from the main welcoming
    // socket, from any of our already-connected sockets, or from the
    // Player's child.

    int max_fd = _welcome_socket;
    fd_set fds;
    FD_ZERO(&fds);
    FD_SET(_welcome_socket, &fds);

    int player_fd = (_active_player != NULL) ? _active_player->get_fd() : -1;
    if (player_fd >= 0) {
      FD_SET(player_fd, &fds);
      max_fd = max(max_fd, player_fd);
    }

    for (ci = _connections.begin(); ci != _connections.end(); ++ci) {
      FD_SET((*ci).first, &fds);
      max_fd = max(max_fd, (*ci).first);
    }

    int result = select(max_fd + 1, &fds, NULL, NULL, NULL);
    if (result < 0) {
      if (errno == EINTR) {
	// If the select terminated because of a signal, it might have
	// been SIGCHLD telling us that one of the player's child
	// process just stopped.  In this case, we need to give the
	// player a chance to update itself.
	for (ci = _connections.begin(); ci != _connections.end(); ++ci) {
	  Player *player = (*ci).second;
	  if (player->is_playing() && player->get_fd() < 0) {
	    // This is the one!
	    player->read_status();
	  }
	}

      } else {
	perror("select");
      }

    } else if (result != 0) {
      // We got some noise on at least one of the sockets.  Which one?
      if (FD_ISSET(_welcome_socket, &fds)) {
	accept_connection();
	result--;
      }

      if (player_fd >= 0 && FD_ISSET(player_fd, &fds)) {
	_active_player->read_status();
	result--;
      }

      for (ci = _connections.begin(); 
	   ci != _connections.end() && result > 0; 
	   ++ci) {
	if (FD_ISSET((*ci).first, &fds)) {
	  (*ci).second->read_user_command();
	  result--;
	}
      }	
    }
  }
}

void Server::
set_active_player(Player *player) {
  // Steal control away from whatever player was active previously.
  if (player != _active_player && _active_player != NULL) {
    _active_player->interrupt();
    // Give it a chance to cleanup and free the sound card.
    sleep(1);
  }
  _active_player = player;
}

const string &Server::
get_sound_device() const {
  return _sound_device;
}

void Server::
close_connection(int socket) {
  Connections::iterator ci = _connections.find(socket);
  if (ci == _connections.end()) {
    cerr << "Unexpected socket passed to close_connection().\n";
  } else {
    // We'll add the Player pointer to our set of connections to be
    // deleted next time around.  We can't delete it right away,
    // because this function was probably called from a method of the
    // to-be-deleted Player, and this circularity would likely cause a
    // seg fault.  Also, we might call this function while we're
    // traversing the _connections list itself, so we can't erase the
    // entry from connections list yet either.
    _closed_connections.insert(*ci);
  }
}

// Tell the server to exit cleanly.
void Server::
clean_exit() {
  _exit_flag = true;
}

// Establish a new connection on a new socket.
void Server::
accept_connection() {
  int newsockfd;

  struct sockaddr_in cli_addr;
  int cli_len = sizeof(cli_addr);
  do {
    newsockfd = accept(_welcome_socket, (struct sockaddr *)&cli_addr, 
		       &cli_len);
  } while (newsockfd < 0 && errno==EINTR);

  if (newsockfd < 0) {
    perror("accept");
  } else {
    _connections[newsockfd] = new Player(newsockfd, this);
  }
}

